/*
 *
 * NNTP Server - See RFC977
 * Jeffrey R. Comstock. - NR0D - Bloomington, Minnesota USA
 * Copyright 1990 Jeffrey R. Comstock, All Rights Reserved.
 * Permission granted for non-commercial copying and use, provided
 * this notice is retained.
 *
 * DB3FL 9107xx: heavily rewritten and bug fixing in file-handling
 * DB3FL 920121: splitted into several files
 *
 */
#include <dos.h>
#include <time.h>
#include <ctype.h>

#include "global.h"
#include "config.h"
#ifdef NNTP
#include "nntp.h"
#include "socket.h"
#include "files.h"
#include "ftp.h"
#include "bm.h"
#ifdef LZW
#include "lzw.h"
#endif

#undef CONTROL				/* (not implemented yet) */
#define	LINE	80

#ifdef LZW
int LzwActive = 1;
#endif

unsigned short Nntpmaxcli = 3;
static unsigned short Nntpsessions = 0;

static char artmsg[] 		= " Article retrieved - ";
static char debug[]		= "100 DEBUG %s\n";
static char help[] 		= "100-ARTICLE  BODY  GROUP   HEAD  HELP IHAVE\n"
					  "100 LAST     LIST  NEWNEWS NEXT  QUIT STAT\n";
/*
static char info[]			= "100 %s Info:\n";
*/
static char xinfo[]		= "100 No info available\n";
static char noactive[]		= "100 No active newsgroups\n";
static char nnversion[]	= "201 %s NNTP version %s ready at %s\n";
static char slave[]    	= "202 SLAVE %s\n";
static char closing[]      = "205 Closing\n";
static char listarticle[]	= "211 %u %u %u%s\n";
static char listgrps[]		= "215 Available newsgroups\n";
static char retrieve[] 	= "220 %u%s%shead and body follow\n";
static char head[]         = "221 %u%s%sHead\n";
static char body[]         = "222 %u%s%sBody\n";
static char statistics[]   = "223 %u%s%sStatistics\n";
static char sepcmd[]       = "223 %u %s%srequest text separately\n";
static char newnews_t[]	= "230 New news by message id follows\n";
static char transok[]		= "235 Thanks\n";

static char sendart[]		= "335 Send article, end with .\n";

static char nogroup[]		= "411 No such newsgroup\n";
static char noselect[]		= "412 No newsgroup selected\n";
static char nonext[]		= "421 No next article\n";
static char noprev[]		= "422 No previous article\n";
static char noart[]		= "430 No such article\n";
static char notwanted[]	= "435 Article not wanted - do not send it\n";
static char transnotok[]	= "437 Article rejected - do not try again\n";

static char badsyntax[]	= "501 Syntax error\n";
static char error[]		= "503 Command not performed\n";
static char lowmem[] 	= "503 System is overloaded\n";
/*
static char nospace[]		= "503 Not enough disk space left\n";
*/

char NEol[] 		= ".\n";

static int SNntp = -1;

/* Start up NNTP receiver service */
int
nntp1(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	struct sockaddr_in lsocket;
	int s;

	if(SNntp != -1)
		return -1;

	psignal(Curproc,0);	/* Don't keep the parser waiting */
	chname(Curproc,"NNTP listener");

	lsocket.sin_family = AF_INET;
	lsocket.sin_addr.s_addr = INADDR_ANY;
	lsocket.sin_port = (argc < 2) ? IPPORT_NNTP : atoi(argv[1]);

	SNntp = socket(AF_INET,SOCK_STREAM,0);
	bind(SNntp,(char *)&lsocket,sizeof(lsocket));
	listen(SNntp,1);

	for(;;){
		if((s = accept(SNntp,NULLCHAR,0)) == -1)
			break;	/* Service is shutting down */

		if(availmem()) {
			/* 503 Spawn a server */
			newproc("NNTP server",4048,nntpserv,s,0,NULL,0);
		} else {
			shutdown(s,1);
		}
	}
	return 0;
}

/* Shutdown NNTP service */
int
nntp0(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	close_s(SNntp);
	SNntp = -1;
	return 0;
}

static int near
strxlen(char *s)
{
	int i = 0;

	while(*s++ != '\0')
		i++;
	return i;
}

/* main message-opening routine
 * returncode: NULLFILE if error; filepointer success */
static FILE * near
open_message(struct nntpsv *mp,FILE *f)
{
	char line[LineLen];

	if(f != NULLFILE)
		fclose(f);

	sprintf(line,"%s/%u",mp->path,mp->pointer);

	if ((f = fopen(line,READ_TEXT)) == NULLFILE)
		usputs(mp->s,noart);
	return f;
}

/* returncode: -1 error; 1 success; 0 no entry */
static int near
get_path(char *group,struct nntpsv *mp)
{
	FILE *f;
	char line[LineLen], *cp;

	if(group == NULLCHAR || mp == NULLNNTPSV
	  || (f = open_file(Pointer,READ_TEXT,mp->s,0)) == NULLFILE)
		return -1;

	group++;

	for (;;) {
		if (fgets(line,LineLen,f) == NULL)
			break;
		if (strcspn(line," ") != strxlen(group))
			continue;
		if (strnicmp(group,line,strxlen(group)) == 0) {
			cp = (strchr(line,' ')) + 1;
			if (mp->path != NULLCHAR)
				xfree(mp->path);
			mp->path = strxdup(cp);
			rip(mp->path);
			fclose(f);
			return 1;
		}
    }
    fclose(f);
	return 0;
}

/* returncode: -1 if error; 1 success; 0 no pointer */
static int near
get_pointer(char *group,struct nntpsv *mp)
{
	FILE *f;
	char line[LineLen], *p;

	if(group == NULLCHAR || mp == NULLNNTPSV
	  || (f = open_file(Active,READ_TEXT,mp->s,0)) == NULLFILE)
		return -1;

	group++;

	for (;;) {
		if (fgets(line,LineLen,f) == NULL)
			break;
		if (strcspn(line," ") != strxlen(group))
			continue;
		if (strnicmp(group,line,strxlen(group))==0) {
			p = strchr(line,' ');
			mp->last = (unsigned)atoi(p);
			mp->first = (unsigned)atoi(strchr(++p,' '));
			mp->pointer = (mp->first > mp->last ) ? 0 : mp->first;
			fclose(f);
			return 1;
		}
    }
    fclose(f);
	return 0;
}

#ifdef CONTROL
static int near
docontrol(FILE *f,struct nntpsv *mp)
{
    struct head *h;

	if(f == NULLFILE || mp == NULLNNTPSV)
		return -1;

	h = (struct head *)mxallocw(sizeof(struct head));

    rewind(f);
	h->subject = h->from = h->reply_to = h->id = NULLCHAR;

	for(;;) {
		if (fgets(mp->buf,LineLen,f) == NULL)
			break;
		if (check_blank(mp->buf))
			break;
		rip(mp->buf);
		if (strncmp(mp->buf,subj,9) == 0)
			h->subject = strxdup(mp->buf);
		if (strncmp(mp->buf,frm,6) == 0)
			h->from = strxdup(mp->buf);
		if (strnicmp(mp->buf,reply_to,10) == 0)
			h->reply_to = strxdup(mp->buf);
		if (strnicmp(mp->buf,msgid,12) == 0)
			h->id = strxdup(strchr(mp->buf,'<'));
	}
	if (h->subject != NULLCHAR)
		if (strncmp(h->subject,"Subject: sendme ",16) == 0)
			dosendme(h);
	if (h->subject != NULLCHAR)
		xfree(h->subject);
	if (h->from != NULLCHAR)
		xfree(h->from);
	if (h->reply_to != NULLCHAR)
		xfree(h->reply_to);
	if (h->id != NULLCHAR)
		xfree(h->id);
	xfree((char *)h);
	return 0;
}
#endif

/* handles incoming newnews-cmd
 * returncode: -1 if error; 0 no groups or bad syntax; 1 success
 */
static int near
donewnews(char *string,struct nntpsv *mp)
{
FILE *f; int ret;

	if((f = temp_file(mp->s,1)) == NULLFILE)
		return -1;

	ret = newnews(string,mp,f);

	switch(ret) {
	case -1:						/* error in "newnews" routine */
		ret = 0;
		usputs(mp->s,badsyntax);
		break;
	case 0:                     	/* no new news */
		usputs(mp->s,newnews_t);
		usputs(mp->s,NEol);
		break;
	default:						/* new news available, send news file */
		usputs(mp->s,newnews_t);
		sendfile(f,mp->s,ASCII_TYPE,0);
		usputs(mp->s,NEol);
		ret = 1;
		break;
	}
	fclose(f);
	return ret;
}

/* change current newsgroup
 * returncode: -1 if error; 0 success; 1 no newsgroup */
static int near
dogroup(struct nntpsv *mp,char *buf)
{
	char *p;
	int er = -1;

	if((p = strchr(buf,' ')) == NULLCHAR)
		return er;

	switch (er = get_path(p,mp)) {
	case 0:
		usputs(mp->s,nogroup);
		er = 1;
		break;
	default:
		p = strchr(buf,' ');
		if (get_pointer(p,mp) == 1) {
			usprintf(mp->s,listarticle,
				mp->last - mp->first + 1,mp->first,mp->last,strchr(buf,' '));
			return 0;
		}
	case -1:
		usputs(mp->s,error);
		break;
	}
	return er;
}


/* checks if newsgroup is selected
 * returncode: -1 no group selected; 0 success */
static int near
check_grp(struct nntpsv *mp)
{
	if(mp == NULLNNTPSV || mp->path == NULLCHAR) {
		usputs(mp->s,noselect);
		return -1;
	}
	return 0;
}

/* get id-number of message
 * returncode: -1 if error; 0 if no message; 1 success */
static int near
get_id(char *bp,struct nntpsv *mp)
{
    FILE *f;
	char tmp[LineLen];

	if ((f = open_message(mp,NULLFILE)) == NULLFILE)
		return 0;

    for (;;) {
		if (fgets(tmp,LineLen,f) == NULL)
			break;
		if (check_blank(tmp))
			break;
		if (strnicmp(tmp,msgid,12)==0) {
			fclose(f);
			strcpy(bp,strchr(tmp,' '));
			rip(bp);
			return 1;
		}
    }
    fclose(f);
    strcpy(bp,"\0");
    return 0;
}

/* gets next news of newsgroup
 * returncode: -1 if error; 0 no news; 1 success */
static int near
get_next(struct nntpsv *mp)
{
	FILE *f;


	for (;;) {
		if (mp->pointer == 0 ) {
			usputs(mp->s,nonext);
			return 0;
		}
		if (++(mp->pointer) > mp->last) {
			mp->pointer--;
			usputs(mp->s,nonext);
			return 0;
		}
		sprintf(mp->buf,"%s/%u",mp->path,mp->pointer);
		if ((f = open_file(mp->buf,READ_TEXT,mp->s,0)) != NULLFILE) {
			fclose(f);
			return 1;
		}
	}
}

/* gets last news of newsgroup
 * returncode: -1 if error; 0 no news; 1 success */
static int near
get_last(struct nntpsv *mp)
{
	FILE *f;

	for (;;) {
		if (mp->pointer == 0) {
			usputs(mp->s,noprev);
			return 0;
		}
		if (--(mp->pointer) < mp->first) {
			mp->pointer++;
			usputs(mp->s,noprev);
			return 0;
		}
		sprintf(mp->buf,"%s/%u",mp->path,mp->pointer);
		if ((f = open_file(mp->buf,READ_TEXT,mp->s,0)) != NULLFILE) {
			fclose(f);
			return 1;
		}
	}
}

static int near
art_ret(struct nntpsv *mp,char flag)
{
	if(get_id(mp->buf,mp) < 1)
		return -1;
	if(flag)
		usprintf(mp->s,retrieve,mp->pointer,mp->buf,artmsg);
	return 0;
}

/* Sends article <message-id> to the client
 * Return value: -1 on error, 1 article sent, 0 no article
 */
int
doarticle(char *buf,struct nntpsv *mp,char flag)
{
	FILE *f;
	char *p, *p2, *holds, line[LineLen];

	if((p = strchr(buf,'<')) == NULLCHAR
	 || (f = fopen(History,READ_TEXT)) == NULLFILE) {
		usputs(mp->s,badsyntax);
		return -1;
	}

	while(fgets(line,LineLen,f) != NULL) {
		if (strstr(line,p) != NULLCHAR) {
			fclose(f);

			p = (strchr(line,' ')) + 14;  	/* point to the first newsgroup */
			p2 = strchr(p,'/');	  			/* point to article number */
			mp->hold_i = mp->pointer;   	/* save current pointer */
			holds = strxdup(mp->path);   	/* save path of current article */
			mp->pointer = atoi((p2 + 1)); 	/* get the article number */

			if (mp->path != NULLCHAR) {
				xfree(mp->path);
				mp->path='\0';
			}

			*p2 = '\0';
			rip(p);

			if (strncmp(p+1,"JUNK",6) == 0) {
				sprintf(mp->buf,"%s",Forward);
				mp->path = strxdup(mp->buf);
			} else
				get_path(p,mp);

			if ((f = open_message(mp,f)) == NULLFILE) {
				return -1;
			}
			if(art_ret(mp,flag) != -1) {
				sendfile(f, mp->s, ASCII_TYPE,0);
				usputs(mp->s,NEol);
			}
			fclose(f);
			xfree(mp->path);
			mp->path = strxdup(holds);
			mp->pointer = mp->hold_i;
			xfree(holds);
			return 1;
	    }

	}
    fclose(f);
	usputs(mp->s,noart);
	return 0;
}

/* checks for a minimum header
 * returncode: -1 if error; 0 complete; 1 not complete */
static int near
garbled(FILE *f)
{
/*
 * The minimum requirement for an incoming article is that it must have
 * a newsgroups: line, a message-id: line, and a blank line.  This is
 * pretty forgiving.
 */
	char line[LineLen];
	int ok = 0;

    rewind(f);
    for(;;) {
		if (fgets(line,LineLen,f) == NULL)
			break;
		if (strnicmp(line,ngrps,12) == 0)
			ok |= 1;
		if (strnicmp(line,msgid,12) == 0)
			ok |= 2;
		if (ok == 3)
			break;
		if (check_blank(line))
			break;
    }
    rewind(f);
	return (ok == 3) ? 0 : 1;
}

/* returncode: -1 if error; 0 success */
static int near
get_article2(struct nntpsv *mp,char *id)
{
	FILE *f;
	int ret = -1;

	if(mp == NULLNNTPSV || id == NULLCHAR
	  || *(id + 1) < ' ' || (f = temp_file(0,1)) == NULLFILE)
		return -1;

	if(recv_file(f,mp->s) != -1)
		if (garbled(f) == 0) {
			mp->id = strxdup(id);
			ret = xfer_article2(f,mp);
		}
	fclose(f);
	usputs(mp->s,ret ? transnotok : transok);
	return ret;
}

static int near
set_pointer(struct nntpsv *mp, char *buf)
{
	char *cp;

	if((cp = strpbrk(buf,"0123456789")) != NULLCHAR) {
		int cnt = atoi(cp);
		if ((cnt > mp->last) || ( cnt < mp->first)) {
			usputs(mp->s,noart);
			return -1;
		}
		mp->pointer = cnt;
	}
	return 0;
}

void
nntpserv(int s,void *unused,void *p)
{
	struct nntpsv *mp;
	FILE *fp;
	long t;
	int cnt;
#ifdef LZW
	int lbits, lmode;
#endif
	char **cmdp, *arg, *cp, *cmd, buf[LineLen];

	/* Command table */
	char *commands[] = {
		"quit",
#define QUIT_CMD	0
		"help",
#define HELP_CMD	1
		"list",
#define LIST_CMD	2
		"group",
#define GROUP_CMD	3
		"debug",
#define DEBUG_CMD	4
		"article",
#define ARTICLE_CMD	5
		"next",
#define NEXT_CMD	6
		"xinfo",
#define XINFO_CMD	7
		"ihave",
#define IHAVE_CMD	8
		"newnews",
#define NEWNEWS_CMD	9
		"head",
#define HEAD_CMD	10
		"body",
#define BODY_CMD	11
		"stat",
#define STAT_CMD	12
		"last",
#define LAST_CMD	13
		"slave",
#define SLAVE_CMD	14
		"xlzw",
#define XLZW_CMD    15
		NULLCHAR
	};

	sockmode(s,SOCK_ASCII);
	sockowner(s,Curproc);		/* We own it now */

	if (!Filecheck)
		if(check_system()) {
			usprintf(s,fatal,"STRUCTURE");
			close_s(s);
			return;
		}

	mp = (struct nntpsv *)mxallocw(sizeof(struct nntpsv));

	mp->s = s;
	log(mp->s,"NNTP open");
	mp->debug = 0;
	mp->id = mp->path = mp->newnews = NULLCHAR;

	time(&t);
	cp = ctime(&t);
	cp[24] = '\0';

	usprintf(mp->s,nnversion,Hostname,Version,cp);

	if(++Nntpsessions > Nntpmaxcli) {
		usputs(mp->s,lowmem);
		goto quit;
	}

loop:
	if ((cnt = recvline(mp->s,buf,LineLen)) == -1) {
		/* He closed on us */
		goto quit;
	}
	rip(buf);
	cmd = buf;

	/* Translate entire buffer to lower case */
	for(cp = cmd; *cp != '\0' && *cp != ' ' ;cp++) {
		*cp = tolower(*cp);
	}
	/* Find command in table; if not present, return syntax error */
	for(cmdp = commands; *cmdp != NULLCHAR; cmdp++)
		if(strnicmp(*cmdp,cmd,strxlen(*cmdp)) == 0)
			break;

	if(*cmdp == NULLCHAR){
		usputs(mp->s,badsyntax);
		goto loop;
	}
	arg = &cmd[strxlen(*cmdp)];

	/* Skip spaces after command */
	while(*arg == ' ')
		arg++;

	/* Execute specific command */
	switch((int)(cmdp - commands)) {
      case XLZW_CMD:
#ifdef LZW
         if (LzwActive)   {
            usprintf(mp->s,transok);
            cp = strchr(arg,' ');
			*cp++ = '\0';
            lbits = atoi(arg); /*get lzwbits*/
            lmode = atoi(cp);
            lzwinit(mp->s,lbits,lmode);
         } else
#endif
			usprintf(mp->s,error);
         break;
	case QUIT_CMD:
		usputs(mp->s,closing);
		goto quit;
	case NEWNEWS_CMD:
		if ((cp = strchr(buf,' ')) == NULLCHAR)
			usputs(mp->s,badsyntax);
		else
			donewnews(++cp,mp);
		break;
	case IHAVE_CMD:
		if ((cp = strchr(buf,'<')) == NULLCHAR)
			usputs(mp->s,badsyntax);
		else if (check_article(cp) == 1)
			usputs(mp->s,notwanted);
		else {
			usputs(mp->s,sendart);
			get_article2(mp,cp);
		}
		break;
	case HELP_CMD:
		usprintf(mp->s,"100-%s - help follows\n",Hostname);
		if ((fp = fopen(Nhelp,READ_TEXT)) != NULLFILE ) {
			sendfile(fp,mp->s,ASCII_TYPE,0);
			fclose(fp);
		} else
			usputs(mp->s,help);
		usputs(mp->s,NEol);
		break;
	case XINFO_CMD:
		if ((fp = fopen(NInfo,READ_TEXT)) != NULLFILE ) {
			usprintf(mp->s,"100-%s - info follows\n",Hostname);
		    sendfile(fp,mp->s,ASCII_TYPE,0);
			fclose(fp);
		} else
			usputs(mp->s,xinfo);
		usputs(mp->s,NEol);
	    break;
	case LIST_CMD:
		if ((fp = open_file(Active,READ_TEXT,mp->s,0)) != NULLFILE) {
			usputs(mp->s,listgrps);
			sendfile(fp,mp->s,ASCII_TYPE,0);
			fclose(fp);
			usputs(mp->s,NEol);
		} else
			usputs(mp->s,noactive);
		break;
	case GROUP_CMD :
		if(dogroup(mp,&buf[0]) == -1)
			usputs(mp->s,badsyntax);
		break;
	case HEAD_CMD :
		if (check_grp(mp))
			break;
		if(set_pointer(mp,buf))
			break;
		if (get_id(mp->buf,mp) == 0)
			break;
		usprintf(mp->s,head,mp->pointer,mp->buf,artmsg);
		if ((fp = open_message(mp,fp)) != NULLFILE) {
			for (;;) {
				if (fgets(mp->buf,LineLen,fp) == NULL)
					break;
				if (check_blank(mp->buf))
					break;
				usputs(mp->s,mp->buf);
			}
			fclose(fp);
		}
		usputs(mp->s,NEol);
		break;
	case BODY_CMD :
		if (check_grp(mp))
			break;
		if(set_pointer(mp,buf))
			break;
		if (get_id(mp->buf,mp) == 0)
			break;
		usprintf(mp->s,body,mp->pointer,mp->buf,artmsg);
		if ((fp = open_message(mp,fp)) != NULLFILE) {
			mp->hold_i = 0;
			for (;;) {
				if (fgets(mp->buf,LineLen,fp) == NULL)
					break;
				if (mp->hold_i)
					usputs(mp->s,mp->buf);
				if (check_blank(mp->buf))
					mp->hold_i = 1;
			}
			fclose(fp);
		}
		usputs(mp->s,NEol);
		break;
	case STAT_CMD :
		if (check_grp(mp))
			break;
		if(set_pointer(mp,buf))
			break;
		if (get_id(mp->buf,mp) == 0)
			break;
		usprintf(mp->s,statistics,mp->pointer,mp->buf,artmsg);
		break;
	case ARTICLE_CMD :
		if (strchr(buf,'<') != NULLCHAR) {
		  doarticle(buf,mp,1);
	      break;
	    }
		if (check_grp(mp))
			break;
		cp = strchr(buf,'e') + 1;

		if ((cnt > 8) && !(check_blank(cp))) {
			if(set_pointer(mp,buf))
				break;
		}
		if ((fp = open_message(mp,fp)) != NULLFILE) {
			art_ret(mp,1);
			sendfile(fp,mp->s,ASCII_TYPE,0);
			fclose(fp);
			usputs(mp->s,NEol);
		}
		break;
	case NEXT_CMD :
		if (check_grp(mp))
		    break;
		if (get_next(mp) == 1) {
			if (get_id(buf,mp) == 1)
				usprintf(mp->s,sepcmd,mp->pointer,&buf[1],artmsg);
		    break;
		}
	    break;
	case LAST_CMD :
		if (check_grp(mp))
			break;
		if (get_last(mp) == 1 ) {
			if (get_id(buf,mp) == 1)
				usprintf(mp->s,sepcmd,mp->pointer,&buf[1],artmsg);
			break;
		}
		break;
	/* This two following cmds currently are not used for much */
	case DEBUG_CMD:
		mp->debug = (mp->debug == 0) ? 1 : 0;
		usprintf(mp->s,debug,(mp->debug == 0) ? "OFF" : "ON");
		break;
	case SLAVE_CMD :
		mp->slave = (mp->slave == 0) ? 1 : 0;
		usprintf(mp->s,slave,(mp->slave == 0) ? "OFF" : "ON");
		break;
	}
	goto loop;

quit:

	log(mp->s,"NNTP close");
	close_s(s);

	Nntpsessions--;

	if(mp->path != NULLCHAR)
		xfree(mp->path);
	if(mp->newnews != NULLCHAR)
		xfree(mp->newnews);
	xfree(mp);
}

/* ---------------------------- SMTP->NNTP-GATE --------------------------- */

int
nnGpost(FILE *data,char *from,struct list *le)
{
	FILE *f;
	struct nntpsv *mp;
	int msgidfound = 0;
	char buf[LineLen], *cp, *cp1;

	if (!Filecheck)
		if(check_system())
			return -1;

	if ((f = temp_file(0,1)) == NULLFILE)
		return -1;

	mp = (struct nntpsv *)mxallocw(sizeof(struct nntpsv));

	/* build path */
	cp = strxdup(from);
	if((cp1 = strpbrk(cp,"@. ")) != NULLCHAR)
		*cp1 = '\0';
	fprintf(f,"%s%s\n",pth,cp);

	/* look for msg-id */
	rewind(data);
	while(fgets(buf,LineLen,data) != NULL)   {
		if(*buf == '\t' || *buf == ' ')
			continue;
		rip(buf);
		if(*buf == '\0')
			break;
		if(htype(buf) == MSGID) {
			msgidfound = 1;
			break;
		}
	}

	/* reorganize header */
	rewind (data);
	while(fgets(buf,LineLen,data) != NULL)   {
		if(*buf == '\t' || *buf == ' ')
			continue;
		rip(buf);
		if(*buf == '\0')
			break;
		switch(htype(buf)) {
		case FROM:
			/* generating from line */
			fprintf(f,"%s%s@%s",frm,cp,Hostname);
			if((cp1 = strpbrk(buf,"<(")) != NULLCHAR)
				fprintf(f," %s",cp1);

			/* generating newsgroups line */
			if((cp1 = strpbrk(&le->val[1],"!@")) != NULLCHAR)
				*cp1 = '\0';
			fprintf(f,"\n%s%s\n",ngrps,&le->val[1]);     /* skip the bang */
			continue;
		case SUBJECT:
			fprintf(f,"%s\n",strxlen(buf) < 10 ? "Subject: (none)" : buf);
			if(msgidfound == 0) {
				sprintf(mp->buf,"<%ld@%s>",get_msgid(),Hostname);
				fprintf(f,"%s%s\n",msgid,mp->buf);
				mp->id = strxdup(mp->buf);
			}
			fprintf(f,"Sender: NNTP@%s\n",Hostname);
			continue;
		case MSGID:
			if(msgidfound != 0) {
				fprintf(f,"%s\n",buf);
				if((cp1 = strchr(buf,'<')) != NULLCHAR) {
					sprintf(mp->buf,"%s",cp1);
					mp->id = strxdup(cp1);
				}
			}
			continue;
		case TO:
		case RECEIVED:
		case SENDER:
		case STATUS:
		case NOHEADER:
			continue;
		}
		fprintf(f,"%s\n",buf);
	}
	xfree(cp);

	fputc('\n',f);

	while(fgets(buf,sizeof(buf),data) != NULL)
		fputs(buf,f);

	fflush(f);
	rewind(f);
	xfer_article2(f,mp);
	fclose(f);
	xfree(mp);
	return 0;
}



#endif /* NNTP */
